@@ -1,10 +1,10 @@
-#!/usr/bin/env python2
+#!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
 """
 yle-dl - download videos from Yle servers
 
-Copyright (C) 2010-2017 Antti Ajanki <antti.ajanki@iki.fi>
+Copyright (C) 2010-2018 Antti Ajanki <antti.ajanki@iki.fi>
 
 This script downloads video and audio streams from Yle Areena
 (https://areena.yle.fi) and Elävä Arkisto
@@ -24,19 +24,21 @@ This script downloads video and audio streams from Yle Areena
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import print_function
+from __future__ import print_function, absolute_import, unicode_literals
 import sys
-import urllib
 import re
-import os.path
-import urlparse
 import codecs
 import logging
 import argparse
-from version import version
-from utils import print_enc
-from downloaders import downloader_factory, StreamFilters, IOContext, \
-    DownloadLimits, BackendFactory, RD_SUCCESS, RD_FAILED
+from future.moves.urllib.parse import urlparse, urlunparse, quote
+from .backends import Backends
+from .downloader import YleDlDownloader
+from .exitcodes import RD_SUCCESS, RD_FAILED
+from .extractors import extractor_factory
+from .io import IOContext, DownloadLimits
+from .streamfilters import StreamFilters
+from .utils import print_enc
+from .version import version
 
 
 def yledl_logger():
@@ -65,6 +67,7 @@ class StreamAction(object):
     PRINT_STREAM_TITLE = 4
     PRINT_EPISODE_PAGES = 5
     PRINT_METADATA = 6
+    DOWNLOAD_SUBTITLES_ONLY = 7
 
 
 def arg_parser():
@@ -75,13 +78,23 @@ def arg_parser():
                     file = sys.stderr
                 print_enc(message, file, False)
 
-    def unicode_arg(bytes):
-        return unicode(bytes, sys.getfilesystemencoding())
+    def to_unicode(s):
+        enc = sys.getfilesystemencoding()
+        try:
+            # Python 2
+            if type(s) == unicode:
+                # default values are unicode even on Python 2
+                return s
+            else:
+                return unicode(s, enc)
+        except NameError:
+            # Python 3
+            return s
 
     description = \
-        (u'yle-dl %s: Download media files from Yle Areena and Elävä Arkisto\n'
-         u'Copyright (C) 2009-2017 Antti Ajanki <antti.ajanki@iki.fi>, '
-         u'license: GPLv3' % version)
+        ('yle-dl %s: Download media files from Yle Areena and Elävä Arkisto\n'
+         'Copyright (C) 2009-2018 Antti Ajanki <antti.ajanki@iki.fi>, '
+         'license: GPLv3' % version)
 
     parser = ArgumentParserEncoded(
         description=description,
@@ -92,20 +105,20 @@ def arg_parser():
 
     io_group = parser.add_argument_group('Input and output')
     url_group = io_group.add_mutually_exclusive_group()
-    url_group.add_argument('url', nargs='?', type=unicode_arg,
-        help=u'Address of an Areena, Elävä Arkisto, or Yle news web page')
+    url_group.add_argument('url', nargs='?', type=to_unicode,
+        help='Address of an Areena, Elävä Arkisto, or Yle news web page')
     url_group.add_argument('-i', metavar='FILENAME', dest='inputfile',
-                           type=unicode_arg,
+                           type=to_unicode,
                            help='Read input URLs to process from the named '
                            'file, one URL per line')
     io_group.add_argument('-o', metavar='FILENAME', dest='outputfile',
-                          type=unicode_arg,
+                          type=to_unicode,
                           help='Save stream to the named file')
     io_group.add_argument('--pipe', action='store_true',
                           help='Dump stream to stdout for piping to media '
                           'player. E.g. "yle-dl --pipe URL | vlc -"')
     io_group.add_argument('--destdir', metavar='DIR',
-                          type=unicode_arg,
+                          type=to_unicode,
                           help='Save files to DIR')
     action_group = io_group.add_mutually_exclusive_group()
     action_group.add_argument('--showurl', action='store_true',
@@ -116,6 +129,8 @@ def arg_parser():
                               help='Print web page for each episode')
     action_group.add_argument('--showmetadata', action='store_true',
                               help='Print metadata about available streams')
+    action_group.add_argument('--subtitlesonly', action='store_true',
+                              help='Download only subtitles, not the video')
     io_group.add_argument('--vfat', action='store_true',
                           help='Output Windows-compatible filenames')
     io_group.add_argument('--resume', action='store_true',
@@ -124,23 +139,23 @@ def arg_parser():
                           help='Maximum bandwidth consumption, '
                           'interger in kB/s')
     io_group.add_argument('--proxy', metavar='URI',
-                          type=unicode_arg,
-                          help='Proxy for downloading streams. '
+                          type=to_unicode,
+                          help='SOCKS or HTTP proxy for downloading streams. '
                           'Example: --proxy socks5://localhost:7777')
     io_group.add_argument('--postprocess', metavar='CMD',
-                          type=unicode_arg,
+                          type=to_unicode,
                           help='Execute the command CMD after a successful '
                           'download. CMD is called with two arguments: '
                           'video, subtitle')
 
     qual_group = parser.add_argument_group('Stream type and quality')
     qual_group.add_argument('--audiolang', metavar='LANG',
-                            type=unicode_arg,
+                            type=to_unicode,
                             choices=['fin', 'swe'], default='',
                             help='Select stream\'s audio language, "fin" or '
                             '"swe"')
     qual_group.add_argument('--sublang', metavar='LANG',
-                            type=unicode_arg,
+                            type=to_unicode,
                             choices=['fin', 'swe', 'smi', 'none', 'all'],
                             help='Download subtitles. LANG is one of "fin", '
                             '"swe", "smi", "none", or "all"')
@@ -149,36 +164,43 @@ def arg_parser():
     qual_group.add_argument('--latestepisode', action='store_true',
                             help='Download the latest episode of a series')
     qual_group.add_argument('--maxbitrate', metavar='RATE',
-                            type=unicode_arg,
+                            type=to_unicode,
                             help='Maximum bitrate stream to download, '
                             'integer in kB/s or "best" or "worst".')
+    qual_group.add_argument('--resolution', metavar='RES',
+                            type=to_unicode,
+                            help='Maximum vertical resolution in pixels, '
+                            'default: highest available resolution')
     qual_group.add_argument('--duration', metavar='S', type=int,
                             help='Record only the first S seconds of '
                             'the stream')
 
     dl_group = parser.add_argument_group('Downloader backends')
     dl_group.add_argument('--backend', metavar='BE',
-                          type=unicode_arg,
-                          default="adobehdsphp,youtubedl",
+                          type=to_unicode,
+                          default="wget,ffmpeg,adobehdsphp,youtubedl,rtmpdump",
                           help='Downloaders that are tried until one of them '
-                          ' succeeds (a comma-separated list).\n'
+                          ' succeeds (a comma-separated list). '
                           'Possible values: '
+                          '"wget", '
+                          '"ffmpeg", '
                           '"adobehdsphp" = AdobeHDS.php, '
-                          '"youtubedl" = youtube-dl')
+                          '"youtubedl" = youtube-dl, '
+                          '"rtmpdump"')
     dl_group.add_argument('--rtmpdump', metavar='PATH',
-                          type=unicode_arg,
+                          type=to_unicode,
                           help='Set path to the rtmpdump binary')
     dl_group.add_argument('--ffmpeg', metavar='PATH',
-                          type=unicode_arg,
+                          type=to_unicode,
                           help='Set path to the ffmpeg binary')
     dl_group.add_argument('--ffprobe', metavar='PATH',
-                          type=unicode_arg,
+                          type=to_unicode,
                           help='Set path to the ffprobe binary')
     dl_group.add_argument('--adobehds', metavar='CMD',
-                          type=unicode_arg, default='',
+                          type=to_unicode, default='',
                           help='Set command for executing AdobeHDS.php')
     dl_group.add_argument('--wget', metavar='PATH',
-                          type=unicode_arg, default='',
+                          type=to_unicode, default='',
                           help='Set path to wget binary')
 
     return parser
@@ -191,19 +213,17 @@ def read_urls_from_file(f):
 
 def encode_url_utf8(url):
     """Encode the path component of url to percent-encoded UTF8."""
-    (scheme, netloc, path, params, query, fragment) = urlparse.urlparse(url)
-
-    path = path.encode('UTF8')
+    (scheme, netloc, path, params, query, fragment) = urlparse(url)
 
     # Assume that the path is already encoded if there seems to be
     # percent encoded entities.
     if re.search(r'%[0-9A-Fa-f]{2}', path) is None:
-        path = urllib.quote(path, '/+')
+        path = quote(path.encode('UTF8'), '/+')
 
-    return urlparse.urlunparse((scheme, netloc, path, params, query, fragment))
+    return urlunparse((scheme, netloc, path, params, query, fragment))
 
 
-def download(url, action, io, stream_filters, backends, postprocess_command):
+def download(url, action, io, stream_filters, postprocess_command):
     """Parse a web page and download the enclosed stream.
 
     url is an Areena, Elävä Arkisto or Yle news web page.
@@ -216,98 +236,77 @@ def download(url, action, io, stream_filters, backends, postprocess_command):
     RD_INCOMPLETE if a stream was downloaded partially but the
     download was interrupted.
     """
-    dl = downloader_factory(url, backends)
-    if not dl:
-        logger.error(u'Unsupported URL %s.' % url)
-        logger.error(u'Is this really a Yle video page?')
+    extractor = extractor_factory(url, stream_filters)
+    if not extractor:
+        logger.error('Unsupported URL %s.' % url)
+        logger.error('Is this really a Yle video page?')
         return RD_FAILED
 
+    clips = extractor.extract(url, stream_filters.latest_only)
+    dl = YleDlDownloader()
+
     if action == StreamAction.PRINT_STREAM_URL:
-        return dl.print_urls(url, stream_filters)
+        print_lines(dl.get_urls(clips, stream_filters))
+        return RD_SUCCESS
     elif action == StreamAction.PRINT_EPISODE_PAGES:
-        return dl.print_episode_pages(url, stream_filters)
+        print_lines(dl.get_episode_pages(clips))
+        return RD_SUCCESS
     elif action == StreamAction.PRINT_STREAM_TITLE:
-        return dl.print_titles(url, io, stream_filters)
+        print_lines(dl.get_titles(clips, io))
+        return RD_SUCCESS
     elif action == StreamAction.PRINT_METADATA:
-        return dl.print_metadata(url, stream_filters)
+        print_lines(dl.get_metadata(clips))
+        return RD_SUCCESS
     elif action == StreamAction.PIPE:
-        return dl.pipe(url, io, stream_filters)
+        return dl.pipe(clips, io, stream_filters)
+    elif action == StreamAction.DOWNLOAD_SUBTITLES_ONLY:
+        dl.download_subtitles(clips, io, stream_filters)
+        return RD_SUCCESS
     else:
-        return dl.download_episodes(url, io, stream_filters,
-                                    postprocess_command)
+        return dl.download_clips(clips, io, stream_filters,
+                                 postprocess_command)
+
+
+def print_lines(lines):
+    for line in lines:
+        print_enc(line)
 
 
 def bitrate_from_arg(arg):
-    if arg == 'best':
-        return sys.maxint
+    if arg is None:
+        return None
+    elif arg == 'best':
+        return None
     elif arg == 'worst':
         return 0
     else:
         try:
             return int(arg)
         except ValueError:
-            logger.warning(u'Invalid bitrate %s, defaulting to best' % arg)
-            arg = sys.maxint
+            logger.warning('Invalid bitrate %s, defaulting to best' % arg)
+            return None
 
 
-def which(program):
-    """Search for program on $PATH and return the full path if found."""
-    # Adapted from http://stackoverflow.com/questions/377017
-    def is_exe(fpath):
-        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
-
-    fpath, fname = os.path.split(program)
-    if fpath:
-        if is_exe(program):
-            return program
-    else:
-        for path in os.environ["PATH"].split(os.pathsep):
-            exe_file = os.path.join(path, program)
-            if is_exe(exe_file):
-                return exe_file
-
-    return None
-
-
-def find_rtmpdump(rtmpdump_arg):
-    binary = rtmpdump_arg
-
-    if not binary:
-        if sys.platform == 'win32':
-            binary = which('rtmpdump.exe')
-        else:
-            binary = which('rtmpdump')
-    if not binary:
-        binary = 'rtmpdump'
-
-    return binary
-
-
-def find_adobehds(adobehds_arg):
-    if adobehds_arg:
-        return adobehds_arg.split(' ')
-    else:
+def resolution_from_arg(arg):
+    if arg is None:
         return None
 
+    if re.match(r'\d+p$', arg):
+        arg = arg[:-1]
 
-def find_ffmpeg(ffmpeg_arg):
-    return ffmpeg_arg or 'ffmpeg'
-
-
-def find_ffprobe(ffprobe_arg):
-    return ffprobe_arg or 'ffprobe'
-
-
-def find_wget(cmd):
-    return cmd or 'wget'
+    try:
+        return int(arg)
+    except ValueError:
+        logger.warning('Invalid resolution: {}'.format(arg))
+        return None
 
 
 ### main program ###
 
 
-def main():
+def main(argv=sys.argv):
     parser = arg_parser()
-    args = parser.parse_args()
+    args = parser.parse_args(argv[1:])
 
     loglevel = logging.DEBUG if args.debug else logging.INFO
     logger.setLevel(loglevel)
@@ -315,10 +314,8 @@ def main():
     excludechars = '\"*/:<>?|' if args.vfat else '*/|'
     dl_limits = DownloadLimits(args.duration, args.ratelimit)
     io = IOContext(args.outputfile, args.destdir, args.resume,
-                   dl_limits, excludechars, args.proxy,
-                   find_rtmpdump(args.rtmpdump),
-                   find_adobehds(args.adobehds), find_ffmpeg(args.ffmpeg),
-                   find_ffprobe(args.ffprobe), find_wget(args.wget))
+                   dl_limits, excludechars, args.proxy, args.rtmpdump,
+                   args.adobehds, args.ffmpeg, args.ffprobe, args.wget)
 
     urls = []
     if args.url:
@@ -341,6 +338,8 @@ def main():
         action = StreamAction.PRINT_METADATA
     elif args.pipe or (args.outputfile == '-'):
         action = StreamAction.PIPE
+    elif args.subtitlesonly:
+        action = StreamAction.DOWNLOAD_SUBTITLES_ONLY
     else:
         action = StreamAction.DOWNLOAD
 
@@ -351,28 +350,25 @@ def main():
                                        StreamAction.PRINT_METADATA]))):
         print_enc(parser.description)
 
-    backends = BackendFactory.parse_backends(args.backend.split(','))
+    backends = Backends.parse_backends(args.backend.split(','))
     if len(backends) == 0:
         sys.exit(RD_FAILED)
 
-    if args.sublang:
-        sublang = args.sublang
-    else:
-        sublang = 'none' if action == StreamAction.PIPE else 'all'
-
-    maxbitrate = bitrate_from_arg(args.maxbitrate or sys.maxint)
+    sublang = args.sublang or 'all'
+    maxbitrate = bitrate_from_arg(args.maxbitrate)
+    maxheight = resolution_from_arg(args.resolution)
     stream_filters = StreamFilters(args.latestepisode, args.audiolang, sublang,
-                                   args.hardsubs, maxbitrate)
+                                   args.hardsubs, maxbitrate, maxheight,
+                                   backends)
     exit_status = RD_SUCCESS
 
     for i, url in enumerate(urls):
         if args.inputfile:
             logger.info('')
-            logger.info(u'Now downloading from URL {}/{}: {}'.format(
+            logger.info('Now downloading from URL {}/{}: {}'.format(
                 i + 1, len(urls), url))
 
-        res = download(url, action, io, stream_filters, backends,
-                       args.postprocess)
+        res = download(url, action, io, stream_filters, args.postprocess)
 
         if res != RD_SUCCESS:
             exit_status = res
